/*
 * echo-server program. Hacked from Echo test suite by
 * <birney@sanger.ac.uk>, ORBit2 udpate by Frank Rehberger
 * <F.Rehberger@xtradyne.de>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <orbit/orbit.h>

#include "echo.h"
#include "echo-skelimpl.c"

static CORBA_ORB global_orb = CORBA_OBJECT_NIL;	/* global orb */
static PortableServer_POA root_poa = CORBA_OBJECT_NIL;	/* root POA */


/* ****************** *
 *  Helper functions  *
 * ****************** */

/* Abort if exception */
static void
abort_if_exception (CORBA_Environment * ev, const char *mesg)
{
  if (ev->_major != CORBA_NO_EXCEPTION)
    {
      g_error ("%s %s", mesg, CORBA_exception_id (ev));
      CORBA_exception_free (ev);
      abort ();
    }
}

/* Writes stringified object reference of @servant to file
 * @filename. If error occures @ev points to exception object on
 * return.
 */
void
export_object_to_file (CORBA_ORB orb,
		       CORBA_Object servant,
		       char *filename, CORBA_Environment * ev)
{
  CORBA_char *objref = NULL;
  FILE *file = NULL;

  /* write objref to file */
  if ((file = fopen (filename, "w")) == NULL)
    g_error ("could not open %s\n", filename);

  /* get objref */
  objref = CORBA_ORB_object_to_string (orb, servant, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  /* print ior to the file */
  fprintf (file, "%s\n", objref);
  fflush (file);

  /* cleaning */
  CORBA_free (objref);
  fclose (file);

  return;
}

/* Is called in case of process signals. it invokes
 * CORBA_ORB_shutdown() function, which will terminate the processes
 * main loop.
 */
static void
server_shutdown (int sig)
{
  CORBA_Environment local_ev[1];
  CORBA_exception_init (local_ev);

  if (global_orb != CORBA_OBJECT_NIL)
    {
      CORBA_ORB_shutdown (global_orb, FALSE, local_ev);
      abort_if_exception (local_ev, "caught exception");
    }
}


/* **************** *
 *  Core functions  *
 * **************** */

/* Inits ORB @orb using @argv arguments for configuration. For each
 * ORBit options consumed from vector @argv the counter of @argc_ptr
 * will be decremented. Signal handler is set to call
 * echo_server_shutdown function in case of SIGINT and SIGTERM
 * signals. If error occures @ev points to exception object on return.
 */
static void
server_init (int *argc_ptr,
	     char *argv[],
	     CORBA_ORB * orb,
	     PortableServer_POA * poa, CORBA_Environment * ev)
{
  PortableServer_POAManager poa_manager = CORBA_OBJECT_NIL;

  CORBA_Environment local_ev[1];
  CORBA_exception_init (local_ev);

  /* init signal handling */
  signal (SIGINT,  server_shutdown);
  signal (SIGTERM, server_shutdown);

  /* create Object Request Broker (ORB) */
  (*orb) = CORBA_ORB_init (argc_ptr, argv, "orbit-local-mt-orb", ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    goto failed_orb;

  (*poa) = (PortableServer_POA)
    CORBA_ORB_resolve_initial_references (*orb, "RootPOA", ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    goto failed_poa;

  poa_manager = PortableServer_POA__get_the_POAManager (*poa, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    goto failed_poamanager;

  PortableServer_POAManager_activate (poa_manager, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    goto failed_activation;

  CORBA_Object_release ((CORBA_Object) poa_manager, ev);
  return;

failed_activation:
failed_poamanager:
  CORBA_Object_release ((CORBA_Object) poa_manager, local_ev);
failed_poa:
  CORBA_ORB_destroy (*orb, local_ev);
failed_orb:
  return;
}

/* Creates servant and registers in context of ORB @orb. The ORB will
 * delegate incoming requests to specific servant object. @return
 * object reference. If error occures @ev points to exception object
 * on return.
 */
static CORBA_Object
server_activate_service (CORBA_ORB orb,
			 PortableServer_POA poa, CORBA_Environment * ev)
{
  ModuleEcho_Echo ref = CORBA_OBJECT_NIL;

  ref = impl_ModuleEcho_Echo__create (poa, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return CORBA_OBJECT_NIL;

  return ref;
}

/* Entering main loop @orb handles incoming request and delegates to
 * servants. If error occures @ev points to exception object on
 * return.
 */
static void
server_run (CORBA_ORB orb, CORBA_Environment * ev)
{
  /* enter main loop until SIGINT or SIGTERM */

  CORBA_ORB_run (orb, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  /* user pressed SIGINT or SIGTERM and in signal handler
   * CORBA_ORB_shutdown(.) has been called
   */
}

/* Releases @servant object and finally destroys @orb. If error
 * occures @ev points to exception object on return.
 */
static void
server_cleanup (CORBA_ORB orb,
		PortableServer_POA poa,
		CORBA_Object ref, CORBA_Environment * ev)
{
  PortableServer_ObjectId *objid = NULL;

  objid = PortableServer_POA_reference_to_id (poa, ref, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  /* Servant: deactivation - will invoke  __fini destructor */
  PortableServer_POA_deactivate_object (poa, objid, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  PortableServer_POA_destroy (poa, TRUE, FALSE, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  CORBA_free (objid);

  CORBA_Object_release ((CORBA_Object) poa, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  CORBA_Object_release (ref, ev);
  if (ev->_major != CORBA_NO_EXCEPTION)
    return;

  /* ORB: tear down the ORB */
  if (orb != CORBA_OBJECT_NIL)
    {
      /* going to destroy orb.. */
      CORBA_ORB_destroy (orb, ev);
      if (ev->_major != CORBA_NO_EXCEPTION)
	return;
    }
}


/* *************** *
 *  Main function  *
 * *************** */

int main (int argc, char *argv[])
{
  CORBA_Object servant = CORBA_OBJECT_NIL;

  CORBA_char filename[] = "echo.ref";

  CORBA_Environment ev[1];
  CORBA_exception_init (ev);

  server_init (&argc, argv, &global_orb, &root_poa, ev);
  abort_if_exception (ev, "failed ORB init");

  servant = server_activate_service (global_orb, root_poa, ev);
  abort_if_exception (ev, "failed activating service");

  printf ("Writing service reference to: %s\n\n", filename);

  export_object_to_file (global_orb, servant, filename, ev);
  abort_if_exception (ev, "failed exporting IOR");

  server_run (global_orb, ev);
  abort_if_exception (ev, "failed entering main loop");

  server_cleanup (global_orb, root_poa, servant, ev);
  abort_if_exception (ev, "failed cleanup");

  return 0;
}
